from django.utils import timezone
from django.test import TestCase, Client
from django.contrib.auth.models import Permission
from knox.models import AuthToken

from authentication.models import ExtendedUser
from files.models import File
from inventory.models import Event, Container, Item

from base64 import b64encode


class ItemTestCase(TestCase):

    def setUp(self):
        super().setUp()
        self.event = Event.objects.create(slug='EVENT', name='Event')
        self.box = Container.objects.create(name='BOX')
        self.user = ExtendedUser.objects.create_user('testuser', 'test', 'test')
        self.user.user_permissions.add(*Permission.objects.all())
        self.token = AuthToken.objects.create(user=self.user)
        self.client = Client(headers={'Authorization': 'Token ' + self.token[1]})

    def test_empty(self):
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'[]')

    def test_members(self):
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(),
                         [{'id': 1, 'description': '1', 'box': 'BOX', 'cid': self.box.id, 'file': None,
                           'returned': False, 'event': self.event.slug}])

    def test_members_with_file(self):
        import base64
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        file = File.objects.create(item=item, data="data:text/plain;base64," + base64.b64encode(b"foo").decode('utf-8'))
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(),
                         [{'id': 1, 'description': '1', 'box': 'BOX', 'cid': self.box.id, 'file': file.hash,
                           'returned': False, 'event': self.event.slug}])

    def test_multi_members(self):
        Item.objects.create(container=self.box, event=self.event, description='1')
        Item.objects.create(container=self.box, event=self.event, description='2')
        Item.objects.create(container=self.box, event=self.event, description='3')
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 3)

    def test_create_item(self):
        response = self.client.post(f'/api/2/{self.event.slug}/item/', {'cid': self.box.id, 'description': '1'})
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json(),
                         {'id': 1, 'description': '1', 'box': 'BOX', 'cid': self.box.id, 'file': None,
                          'returned': False, 'event': self.event.slug})
        self.assertEqual(len(Item.objects.all()), 1)
        self.assertEqual(Item.objects.all()[0].id, 1)
        self.assertEqual(Item.objects.all()[0].description, '1')
        self.assertEqual(Item.objects.all()[0].container.id, self.box.id)

    def test_create_item_without_container(self):
        response = self.client.post(f'/api/2/{self.event.slug}/item/', {'description': '1'})
        self.assertEqual(response.status_code, 400)

    def test_create_item_without_description(self):
        response = self.client.post(f'/api/2/{self.event.slug}/item/', {'cid': self.box.id})
        self.assertEqual(response.status_code, 400)

    def test_create_item_with_file(self):
        import base64
        response = self.client.post(f'/api/2/{self.event.slug}/item/',
                                    {'cid': self.box.id, 'description': '1',
                                     'dataImage': "data:text/plain;base64," + base64.b64encode(b"foo").decode(
                                         'utf-8')}, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()['id'], 1)
        self.assertEqual(response.json()['description'], '1')
        self.assertEqual(response.json()['box'], 'BOX')
        self.assertEqual(response.json()['id'], self.box.id)
        self.assertEqual(len(response.json()['file']), 64)
        self.assertEqual(len(Item.objects.all()), 1)
        self.assertEqual(Item.objects.all()[0].id, 1)
        self.assertEqual(Item.objects.all()[0].description, '1')
        self.assertEqual(Item.objects.all()[0].container.id, self.box.id)
        self.assertEqual(len(File.objects.all()), 1)

    def test_update_item(self):
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        response = self.client.patch(f'/api/2/{self.event.slug}/item/{item.id}/', {'description': '2'},
                                   content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(),
                         {'id': 1, 'description': '2', 'box': 'BOX', 'cid': self.box.id, 'file': None,
                          'returned': False, 'event': self.event.slug})
        self.assertEqual(len(Item.objects.all()), 1)
        self.assertEqual(Item.objects.all()[0].id, 1)
        self.assertEqual(Item.objects.all()[0].description, '2')
        self.assertEqual(Item.objects.all()[0].container.id, self.box.id)

    def test_update_item_with_file(self):
        import base64
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        response = self.client.patch(f'/api/2/{self.event.slug}/item/{item.id}/',
                                   {'description': '2',
                                    'dataImage': "data:text/plain;base64," + base64.b64encode(b"foo").decode('utf-8')},
                                   content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['id'], 1)
        self.assertEqual(response.json()['description'], '2')
        self.assertEqual(response.json()['box'], 'BOX')
        self.assertEqual(response.json()['id'], self.box.id)
        self.assertEqual(len(response.json()['file']), 64)
        self.assertEqual(len(Item.objects.all()), 1)
        self.assertEqual(Item.objects.all()[0].id, 1)
        self.assertEqual(Item.objects.all()[0].description, '2')
        self.assertEqual(Item.objects.all()[0].container.id, self.box.id)
        self.assertEqual(len(File.objects.all()), 1)

    def test_delete_item(self):
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        Item.objects.create(container=self.box, event=self.event, description='2')
        self.assertEqual(len(Item.objects.all()), 2)
        response = self.client.delete(f'/api/2/{self.event.slug}/item/{item.id}/')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(len(Item.objects.all()), 1)

    def test_delete_item2(self):
        Item.objects.create(container=self.box, event=self.event, description='1')
        item2 = Item.objects.create(container=self.box, event=self.event, description='2')
        self.assertEqual(len(Item.objects.all()), 2)
        response = self.client.delete(f'/api/2/{self.event.slug}/item/{item2.id}/')
        self.assertEqual(response.status_code, 204)
        self.assertEqual(len(Item.objects.all()), 1)
        item3 = Item.objects.create(container=self.box, event=self.event, description='3')
        self.assertEqual(item3.id, 3)
        self.assertEqual(len(Item.objects.all()), 2)

    def test_item_count(self):
        Item.objects.create(container=self.box, event=self.event, description='1')
        Item.objects.create(container=self.box, event=self.event, description='2')
        response = self.client.get('/api/2/boxes/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual(response.json()[0]['itemCount'], 2)

    def test_item_nonexistent(self):
        response = self.client.get(f'/api/2/NOEVENT/item/')
        self.assertEqual(response.status_code, 404)

    def test_item_return(self):
        item = Item.objects.create(container=self.box, event=self.event, description='1')
        self.assertEqual(item.returned_at, None)
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        response = self.client.patch(f'/api/2/{self.event.slug}/item/{item.id}/', {'returned': True},
                                     content_type='application/json')
        self.assertEqual(response.status_code, 200)
        item.refresh_from_db()
        self.assertNotEqual(item.returned_at, None)
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

    def test_item_show_not_returned(self):
        item1 = Item.objects.create(container=self.box, event=self.event, description='1')
        item2 = Item.objects.create(container=self.box, event=self.event, description='2')
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 2)
        item2.returned_at = timezone.now()
        item2.save()
        response = self.client.get(f'/api/2/{self.event.slug}/item/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual(response.json()[0]['id'], item1.id)


class ItemSearchTestCase(TestCase):

    def setUp(self):
        super().setUp()
        self.event = Event.objects.create(slug='EVENT', name='Event')
        self.box = Container.objects.create(name='BOX')
        self.user = ExtendedUser.objects.create_user('testuser', 'test', 'test')
        self.user.user_permissions.add(*Permission.objects.all())
        self.token = AuthToken.objects.create(user=self.user)
        self.client = Client(headers={'Authorization': 'Token ' + self.token[1]})
        self.item1 = Item.objects.create(container=self.box, event=self.event, description='abc def')
        self.item2 = Item.objects.create(container=self.box, event=self.event, description='def ghi')
        self.item3 = Item.objects.create(container=self.box, event=self.event, description='jkl mno pqr')
        self.item4 = Item.objects.create(container=self.box, event=self.event, description='stu vwx')

    def test_search(self):
        search_query = b64encode(b'abc').decode('utf-8')
        response = self.client.get(f'/api/2/{self.event.slug}/items/{search_query}/')
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(response.json()))
        self.assertEqual(self.item1.id, response.json()[0]['id'])
        self.assertEqual('abc def', response.json()[0]['description'])
        self.assertEqual('BOX', response.json()[0]['box'])
        self.assertEqual(self.box.id, response.json()[0]['cid'])
        self.assertEqual(1, response.json()[0]['search_score'])

    def test_search2(self):
        search_query = b64encode(b'def').decode('utf-8')
        response = self.client.get(f'/api/2/{self.event.slug}/items/{search_query}/')
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, len(response.json()))
        self.assertEqual(self.item1.id, response.json()[0]['id'])
        self.assertEqual('abc def', response.json()[0]['description'])
        self.assertEqual('BOX', response.json()[0]['box'])
        self.assertEqual(self.box.id, response.json()[0]['id'])
        self.assertEqual(1, response.json()[0]['search_score'])
        self.assertEqual(self.item2.id, response.json()[1]['id'])
        self.assertEqual('def ghi', response.json()[1]['description'])
        self.assertEqual('BOX', response.json()[1]['box'])
        self.assertEqual(self.box.id, response.json()[1]['cid'])
        self.assertEqual(1, response.json()[0]['search_score'])

    def test_search3(self):
        search_query = b64encode(b'jkl').decode('utf-8')
        response = self.client.get(f'/api/2/{self.event.slug}/items/{search_query}/')
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, len(response.json()))
        self.assertEqual(self.item3.id, response.json()[0]['id'])
        self.assertEqual('jkl mno pqr', response.json()[0]['description'])
        self.assertEqual('BOX', response.json()[0]['box'])
        self.assertEqual(self.box.id, response.json()[0]['cid'])
        self.assertEqual(1, response.json()[0]['search_score'])

    def test_search4(self):
        search_query = b64encode(b'abc def').decode('utf-8')
        response = self.client.get(f'/api/2/{self.event.slug}/items/{search_query}/')
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, len(response.json()))
        self.assertEqual(self.item1.id, response.json()[0]['id'])
        self.assertEqual('abc def', response.json()[0]['description'])
        self.assertEqual('BOX', response.json()[0]['box'])
        self.assertEqual(self.box.id, response.json()[0]['id'])
        self.assertEqual(2, response.json()[0]['search_score'])
        self.assertEqual(self.item2.id, response.json()[1]['id'])
        self.assertEqual('def ghi', response.json()[1]['description'])
        self.assertEqual('BOX', response.json()[1]['box'])
        self.assertEqual(self.box.id, response.json()[1]['cid'])
        self.assertEqual(1, response.json()[1]['search_score'])

